use io;

/* Read instance data */
function input() {
    local usage = "Usage: localsolver car_sequencing.lsp "
            + "inFileName=inputFile [solFileName=outputFile] [lsTimeLimit=timeLimit]";

    if (inFileName == nil) throw usage;
    local inFile = io.openRead(inFileName);
    nbPositions = inFile.readInt();
    nbOptions = inFile.readInt();
    nbClasses = inFile.readInt();

    maxCarsPerWindow[o in 0...nbOptions] = inFile.readInt();
    windowSize[o in 0...nbOptions] = inFile.readInt();

    local pos = 0;
    for [c in 0...nbClasses] {
        inFile.readInt(); // Note: index of class is read but not used
        nbCars[c] = inFile.readInt();
        options[c][o in 0...nbOptions] = inFile.readInt();
        initialSequence[p in pos...pos+nbCars[c]] = c;
        pos += nbCars[c];
    }
}

/* Declare the optimization model */
function model() {
    // sequence[i] = j if class initially placed on position j is produced on position i
    sequence <- list(nbPositions);
    
    // sequence is a permutation of the initial production plan, all indexes must
    // appear exactly once
    constraint partition(sequence);

    // Number of cars with option o in each window
    nbCarsWindows[o in 0...nbOptions][j in 0...nbPositions-windowSize[o]+1]
            <- sum[k in 0...windowSize[o]](options[initialSequence[sequence[j + k]]][o]);

    // Number of violations of option o capacity in each window
    nbViolationsWindows[o in 0...nbOptions][p in 0...nbPositions-windowSize[o]+1]
            <- max(nbCarsWindows[o][p] - maxCarsPerWindow[o], 0);

    // Minimize the sum of violations for all options and all windows
    totalViolations <- sum[o in 0...nbOptions][p in 0...nbPositions-windowSize[o]+1](
            nbViolationsWindows[o][p]);
    minimize totalViolations;
}

/* Parametrize the solver */
function param() {
    // Set the initial solution
    sequence.value.clear();
    for [p in 0...nbPositions]
        sequence.value.add(p);
    if (lsTimeLimit == nil) lsTimeLimit = 60;
}

/* Write the solution in a file with the following format: 
 * - 1st line: value of the objective;
 * - 2nd line: for each position p, index of class at positions p. */
function output() {
    if (solFileName == nil) return;
    local solFile = io.openWrite(solFileName);
    solFile.println(totalViolations.value);
    local listSolution = sequence.value;
    for [p in 0...nbPositions]
        solFile.print(initialSequence[listSolution[p]], " ");
    solFile.println();
}

